# coding: utf-8

from supervisely.sly_logger import logger, ServiceType, EventType, add_logger_handler, \
    add_default_logging_into_file, get_task_logger, change_formatters_default_values, LOGGING_LEVELS

from supervisely.function_wrapper import main_wrapper, function_wrapper, catch_silently, function_wrapper_nofail, \
    function_wrapper_external_logger

from supervisely.io import fs
from supervisely.io import env
from supervisely.io import json
from supervisely.io import network_exceptions
from supervisely.io.fs_cache import FileCache

from supervisely.imaging import image
# legacy
# import supervisely.imaging.video as imagevideo
from supervisely.imaging import color

from supervisely.task.paths import TaskPaths

from supervisely.task.progress import epoch_float, Progress, report_import_finished, report_dtl_finished, \
    report_dtl_verification_finished, \
    report_metrics_training, report_metrics_validation, report_inference_finished

from supervisely.project.project import Project, OpenMode, download_project, read_single_project, upload_project, \
    Dataset
from supervisely.project.project_meta import ProjectMeta

from supervisely.annotation.annotation import ANN_EXT, Annotation
from supervisely.annotation.label import Label
from supervisely.annotation.obj_class import ObjClass, ObjClassJsonFields
from supervisely.annotation.obj_class_collection import ObjClassCollection
from supervisely.annotation.tag_meta import TagMeta, TagValueType, TagApplicableTo
from supervisely.annotation.tag import Tag
from supervisely.annotation.tag_collection import TagCollection
from supervisely.annotation.tag_meta_collection import TagMetaCollection

from supervisely.geometry.bitmap import Bitmap
from supervisely.geometry.cuboid import Cuboid
from supervisely.geometry.point import Point
from supervisely.geometry.point_location import PointLocation
from supervisely.geometry.polygon import Polygon
from supervisely.geometry.polyline import Polyline
from supervisely.geometry.rectangle import Rectangle
from supervisely.geometry.any_geometry import AnyGeometry
from supervisely.geometry.multichannel_bitmap import MultichannelBitmap

from supervisely.geometry.helpers import geometry_to_bitmap
from supervisely.geometry.helpers import deserialize_geometry

from supervisely.export.pascal_voc import save_project_as_pascal_voc_detection

from supervisely.metric.metric_base import MetricsBase
from supervisely.metric.projects_applier import MetricProjectsApplier

from supervisely.metric.iou_metric import IoUMetric
from supervisely.metric.confusion_matrix_metric import ConfusionMatrixMetric
from supervisely.metric.precision_recall_metric import PrecisionRecallMetric
from supervisely.metric.classification_metrics import ClassificationMetrics
from supervisely.metric.map_metric import MAPMetric

from supervisely.worker_api.agent_api import AgentAPI
from supervisely.worker_api.chunking import ChunkSplitter, ChunkedFileWriter, ChunkedFileReader
import supervisely.worker_proto.worker_api_pb2 as api_proto

from supervisely.api.api import Api
from supervisely.api import api
from supervisely.api.task_api import WaitingTimeExceeded
from supervisely.project.project_type import ProjectType
from supervisely.api.report_api import NotificationType

from supervisely._utils import rand_str, batched, get_bytes_hash, generate_names, ENTERPRISE, COMMUNITY, _dprint, \
    take_with_default, get_string_hash
from supervisely.tiny_timer import TinyTimer

from supervisely.aug import aug

from supervisely.video_annotation.video_annotation import VideoAnnotation
from supervisely.video_annotation.video_object import VideoObject
from supervisely.video_annotation.video_object_collection import VideoObjectCollection
from supervisely.video_annotation.video_figure import VideoFigure
from supervisely.video_annotation.frame import Frame
from supervisely.video_annotation.frame_collection import FrameCollection
from supervisely.project.video_project import VideoDataset, VideoProject, download_video_project, upload_video_project
from supervisely.video import video

import supervisely.labeling_jobs.utils as lj

from supervisely.pointcloud import pointcloud
from supervisely.pointcloud_annotation.pointcloud_annotation import PointcloudAnnotation
from supervisely.pointcloud_annotation.pointcloud_episode_annotation import PointcloudEpisodeAnnotation
from supervisely.pointcloud_annotation.pointcloud_object import PointcloudObject
from supervisely.pointcloud_annotation.pointcloud_figure import PointcloudFigure
from supervisely.project.pointcloud_project import PointcloudDataset, PointcloudProject, download_pointcloud_project
from supervisely.project.pointcloud_episode_project import PointcloudEpisodeDataset, PointcloudEpisodeProject, download_pointcloud_episode_project

from supervisely.pyscripts_utils import utils as ps
from supervisely.io import docker_utils
# import supervisely.app as app
from supervisely.app.v1.app_service import AppService

from supervisely.decorators.profile import timeit
from supervisely.decorators.profile import update_fields
from supervisely.decorators.inference import process_image_roi

import supervisely.script as script
from supervisely.user.user import UserRoleName
from supervisely.io import github_utils as git

from supervisely.aug import imgaug_utils



# coding: utf-8

from supervisely.api.module_api import ModuleApi
from supervisely.api.module_api import ApiField
from supervisely.video_annotation.key_id_map import KeyIdMap


class ObjectClassApi(ModuleApi):
    @staticmethod
    def info_sequence():
        return [ApiField.ID,
                ApiField.NAME,
                ApiField.DESCRIPTION,
                ApiField.SHAPE,
                ApiField.COLOR,
                ApiField.SETTINGS,
                ApiField.CREATED_AT,
                ApiField.UPDATED_AT
                ]

    @staticmethod
    def info_tuple_name():
        return 'ObjectClassInfo'

    def get_list(self, project_id, filters=None):
        '''
        :param project_id: int
        :param filters: list
        :return: List the object classes from the given project
        '''
        return self.get_list_all_pages('advanced.object_classes.list',  {ApiField.PROJECT_ID: project_id, "filter": filters or []})

    def get_name_to_id_map(self, project_id):
        '''
        :param project_id: int
        :return: dictionary object class name -> object class id
        '''
        objects_infos = self.get_list(project_id)
        return {object_info.name: object_info.id for object_info in objects_infos}

    # def _object_classes_to_json(self, object_classes: KeyIndexedCollection, objclasses_name_id_map=None, project_id=None):
    #     pass #@TODO: implement
    #     # if objclasses_name_id_map is None and project_id is None:
    #     #     raise RuntimeError("Impossible to get ids for projectTags")
    #     # if objclasses_name_id_map is None:
    #     #     objclasses_name_id_map = self.get_name_to_id_map(project_id)
    #     # tags_json = []
    #     # for tag in tags:
    #     #     tag_json = tag.to_json()
    #     #     tag_json[ApiField.TAG_ID] = tag_name_id_map[tag.name]
    #     #     tags_json.append(tag_json)
    #     # return tags_json
    # 
    # def append_to_video(self, video_id, tags: KeyIndexedCollection, key_id_map: KeyIdMap = None):
    #     if len(tags) == 0:
    #         return []
    #     video_info = self._api.video.get_info_by_id(video_id)
    #     tags_json = self._tags_to_json(tags, project_id=video_info.project_id)
    #     ids = self.append_to_video_json(video_id, tags_json)
    #     KeyIdMap.add_tags_to(key_id_map, [tag.key() for tag in tags], ids)
    #     return ids
    # 
    # def append_to_video_json(self, video_id, tags_json):
    #     if len(tags_json) == 0:
    #         return
    #     response = self._api.post('videos.tags.bulk.add', {ApiField.VIDEO_ID: video_id, ApiField.TAGS: tags_json})
    #     ids = [obj[ApiField.ID] for obj in response.json()]
    #     return ids





# coding: utf-8

from copy import deepcopy
import cv2
import numpy as np

from supervisely.geometry.constants import EXTERIOR, INTERIOR, POINTS, GEOMETRY_SHAPE, GEOMETRY_TYPE
from supervisely.geometry.geometry import Geometry
from supervisely.geometry.point_location import PointLocation, points_to_row_col_list
from supervisely.geometry.rectangle import Rectangle


class VectorGeometry(Geometry):
    '''
    This is a base class for creating and using VectorGeometry objects for Labels
    '''
    def __init__(self, exterior, interior,
                 sly_id=None, class_id=None, labeler_login=None, updated_at=None, created_at=None):
        '''
        :param exterior: list of PointLocation objects
        :param interior: list of PointLocation objects
        '''
        if not (isinstance(exterior, list) and all(isinstance(p, PointLocation) for p in exterior)):
            raise TypeError('Argument "exterior" must be list of "PointLocation" objects!')

        if not isinstance(interior, list) or \
            not all(isinstance(c, list) for c in interior) or \
                not all(isinstance(p, PointLocation) for c in interior for p in c):
            raise TypeError('Argument "interior" must be list of list of "PointLocation" objects!')

        self._exterior = deepcopy(exterior)
        self._interior = deepcopy(interior)
        super().__init__(sly_id=sly_id, class_id=class_id, labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)

    def to_json(self):
        '''
        The function from_json convert VectorGeometry class object to json format
        :return: VectorGeometry class object in json format
        '''
        packed_obj = {
            POINTS: {
                EXTERIOR: points_to_row_col_list(self._exterior, flip_row_col_order=True),
                INTERIOR: [points_to_row_col_list(i, flip_row_col_order=True) for i in self._interior]
            },
            GEOMETRY_SHAPE: self.geometry_name(),
            GEOMETRY_TYPE: self.geometry_name(),
        }
        self._add_creation_info(packed_obj)
        return packed_obj

    @property
    def exterior(self):
        return deepcopy(self._exterior)

    @property
    def exterior_np(self):
        '''
        The function exterior_np convert exterior attribute(list of PointLocation objects) to numpy array
        :return: numpy array
        '''
        return np.array(points_to_row_col_list(self._exterior), dtype=np.int64)

    @property
    def interior(self):
        return deepcopy(self._interior)

    @property
    def interior_np(self):
        '''
        The function interior_np convert interior attribute(list of PointLocation objects) to numpy array
        :return: numpy array
        '''
        return [np.array(points_to_row_col_list(i), dtype=np.int64) for i in self._interior]

    def _transform(self, transform_fn):
        result = deepcopy(self)
        result._exterior = [transform_fn(p) for p in self._exterior]
        result._interior = [[transform_fn(p) for p in i] for i in self._interior]
        return result

    def resize(self, in_size, out_size):
        '''
        Resize the current VectorGeometry to match a certain size
        :param in_size: input image size
        :param out_size: output image size
        :return: VectorGeometry class object
        '''
        return self._transform(lambda p: p.resize(in_size, out_size))

    def scale(self, factor):
        '''
        The function scale change scale of the current VectorGeometry object with a given factor
        :param factor: float scale parameter
        :return: VectorGeometry class object
        '''
        return self._transform(lambda p: p.scale(factor))

    def translate(self, drow, dcol):
        '''
        The function translate shifts the VectorGeometry object by a certain number of pixels and return the copy of the current VectorGeometry object
        :param drow: horizontal shift
        :param dcol: vertical shift
        :return: VectorGeometry class object
        '''
        return self._transform(lambda p: p.translate(drow, dcol))

    def rotate(self, rotator):
        '''
        The function rotate VectorGeometry with a given rotator(ImageRotator class object contain size of image and angle to rotate)
        :param rotator: ImageRotator class object
        :return: VectorGeometry class object
        '''
        return self._transform(lambda p: p.rotate(rotator))

    def fliplr(self, img_size):
        '''
        The function fliplr the current VectorGeometry object geometry in horizontal
        :param img_size: size of the image
        :return: VectorGeometry class object
        '''
        return self._transform(lambda p: p.fliplr(img_size))

    def flipud(self, img_size):
        '''
        The function flipud the current VectorGeometry object geometry in vertical
        :param img_size: size of the image
        :return: VectorGeometry class object
        '''
        return self._transform(lambda p: p.flipud(img_size))

    def to_bbox(self):
        '''
        The function to_bbox create Rectangle class object from current VectorGeometry class object
        :return: Rectangle class object
        '''
        exterior_np = self.exterior_np
        rows, cols = exterior_np[:, 0], exterior_np[:, 1]
        return Rectangle(top=round(min(rows).item()), left=round(min(cols).item()), bottom=round(max(rows).item()),
                         right=round(max(cols).item()))

    def _draw_impl(self, bitmap, color, thickness=1, config=None):
        """
        :param bitmap: np.ndarray
        :param color: [R, G, B]
        :param thickness: used only in Polyline and Point
        """
        self._draw_contour_impl(bitmap, color, thickness, config=config)

    def _draw_contour_impl(self, bitmap, color, thickness=1, config=None):
        """ Draws the figure contour on a given bitmap canvas
        :param bitmap: np.ndarray
        :param color: [R, G, B]
        :param thickness: (int)
        """
        raise NotImplementedError()

    @staticmethod
    def _approx_ring_dp(ring, epsilon, closed):
        new_ring = cv2.approxPolyDP(ring.astype(np.int32), epsilon, closed)
        new_ring = np.squeeze(new_ring, 1)
        if len(new_ring) < 3 and closed:
            new_ring = ring.astype(np.int32)
        return new_ring

    def approx_dp(self, epsilon):
        raise NotImplementedError()



# coding: utf-8

import urllib
from supervisely.api.module_api import ApiField, ModuleApi, UpdateableModule, RemoveableModuleApi


class DatasetApi(UpdateableModule, RemoveableModuleApi):
    '''
    This is a class for creating and using DatasetApi objects. Here is where your labeled and unlabeled images and other
    files live. There is no more levels: images or videos are directly attached to a dataset. Dataset is a unit of work.
    All images or videos are directly attached to a dataset. A dataset is some sort of data folder with stuff to annotate.
    '''
    @staticmethod
    def info_sequence():
        return [ApiField.ID,
                ApiField.NAME,
                ApiField.DESCRIPTION,
                ApiField.SIZE,
                ApiField.PROJECT_ID,
                ApiField.IMAGES_COUNT,
                ApiField.ITEMS_COUNT,
                ApiField.CREATED_AT,
                ApiField.UPDATED_AT,
                ApiField.REFERENCE_IMAGE_URL]

    @staticmethod
    def info_tuple_name():
        return 'DatasetInfo'

    def __init__(self, api):
        ModuleApi.__init__(self, api)
        UpdateableModule.__init__(self, api)

    def get_list(self, project_id, filters=None):
        '''
        :param project_id: int
        :param filters: list
        :return: list all the datasets for a given project
        '''
        return self.get_list_all_pages('datasets.list',  {ApiField.PROJECT_ID: project_id, ApiField.FILTER: filters or []})

    def get_info_by_id(self, id):
        '''
        :param id: int
        :return: dataset metadata by numeric id
        '''
        return self._get_info_by_id(id, 'datasets.info')

    def create(self, project_id, name, description="", change_name_if_conflict=False):
        '''
        Create dataset with given name in project with given id
        :param project_id: int
        :param name: str
        :param description: str
        :param change_name_if_conflict: bool
        :return: created project dataset metadata
        '''
        effective_name = self._get_effective_new_name(
            parent_id=project_id, name=name, change_name_if_conflict=change_name_if_conflict)
        response = self._api.post('datasets.add', {ApiField.PROJECT_ID: project_id,
                                                   ApiField.NAME: effective_name,
                                                   ApiField.DESCRIPTION: description})
        return self._convert_json_info(response.json())

    def get_or_create(self, project_id, name, description=""):
        '''
        Check if dataset with given name exist in project, if not it create dataset with given name in project
        :param project_id: int
        :param name: str
        :param description: str
        :return: dataset metadata
        '''
        dataset_info = self.get_info_by_name(project_id, name)
        if dataset_info is None:
            dataset_info = self.create(project_id, name, description=description)
        return dataset_info

    def _get_update_method(self):
        return 'datasets.editInfo'

    def _remove_api_method_name(self):
        return 'datasets.remove'

    def copy_batch(self, dst_project_id, ids, new_names=None, change_name_if_conflict=False, with_annotations=False):
        '''
        Copy given datasets in destination project
        :param dst_project_id: int
        :param ids: list of integers
        :param new_names: list of str (if new_names not None and lengh of ids list != lengh of new_names list generate error)
        :param change_name_if_conflict: bool
        :param with_annotations: bool
        :return: list of datasets (DatasetApi class objects)
        '''
        if new_names is not None and len(ids) != len(new_names):
            raise RuntimeError('Can not match "ids" and "new_names" lists, len(ids) != len(new_names)')

        new_datasets = []
        for idx, dataset_id in enumerate(ids):
            dataset = self.get_info_by_id(dataset_id)
            new_dataset_name = dataset.name
            if new_names is not None:
                new_dataset_name = new_names[idx]
            src_images = self._api.image.get_list(dataset.id)
            src_image_ids = [image.id for image in src_images]
            new_dataset = self._api.dataset.create(dst_project_id, new_dataset_name, dataset.description,
                                                   change_name_if_conflict=change_name_if_conflict)
            self._api.image.copy_batch(new_dataset.id, src_image_ids, change_name_if_conflict, with_annotations)
            new_datasets.append(new_dataset)
        return new_datasets

    def copy(self, dst_project_id, id, new_name=None, change_name_if_conflict=False, with_annotations=False):
        '''
        Copy given dataset in destination project
        :param dst_project_id: int
        :param id: int
        :param new_name: str
        :param change_name_if_conflict: bool
        :param with_annotations: bool
        :return: DatasetApi class object
        '''
        new_datasets = self.copy_batch(dst_project_id, [id], [new_name], change_name_if_conflict, with_annotations)
        if len(new_datasets) == 0:
            return None
        return new_datasets[0]

    def move_batch(self, dst_project_id, ids, new_names, change_name_if_conflict=False, with_annotations=False):
        '''
        Moves given datasets in destination project
        :param dst_project_id: int
        :param ids: list of integers
        :param new_names: list of str (if new_names not None and lengh of ids list != lengh of new_names list generate error)
        :param change_name_if_conflict: bool
        :param with_annotations: bool
        :return: list of datasets (DatasetApi class objects)
        '''
        new_datasets = self.copy_batch(dst_project_id, ids, new_names, change_name_if_conflict, with_annotations)
        self.remove_batch(ids)
        return new_datasets

    def move(self, dst_project_id, id, new_name, change_name_if_conflict=False, with_annotations=False):
        '''
        Move given dataset in destination project
        :param dst_project_id: int
        :param id: int
        :param new_name: str
        :param change_name_if_conflict: bool
        :param with_annotations: bool
        :return: DatasetApi class object
        '''
        new_dataset = self.copy(dst_project_id, id, new_name, change_name_if_conflict, with_annotations)
        self.remove(id)
        return new_dataset

    def _convert_json_info(self, info: dict, skip_missing=True):
        res = super()._convert_json_info(info, skip_missing=skip_missing)
        if res.reference_image_url is not None:
            res = res._replace(reference_image_url=urllib.parse.urljoin(self._api.server_address, res.reference_image_url))
        if res.items_count is None:
            res = res._replace(items_count=res.images_count)
        return res




from typing import Literal
import markupsafe

from supervisely.app import DataJson
from supervisely.app.widgets import Widget

INFO = "info"
WARNING = "warning"
ERROR = "error"

from pathlib import Path
from jinja2 import Environment
import jinja2


class NotificationBox(Widget):
    def __init__(
        self,
        title: str = None,
        description: str = None,
        box_type: Literal["info", "warning", "error"] = WARNING,
        widget_id: str = None,
    ):
        self.title = title
        self.description = description
        if self.title is None and self.description is None:
            raise ValueError(
                "Both title and description can not be None at the same time"
            )
        self.box_type = box_type
        self.icon = "zmdi-alert-triangle"  # @TODO: get by box type
        if self.box_type != WARNING:
            raise ValueError(
                f"Only {WARNING} type is supported. Other types {[INFO, WARNING, ERROR]} will be supported later"
            )

        super().__init__(widget_id=widget_id, file_path=__file__)

    def init_data(self):
        return {"title": self.title, "description": self.description, "icon": self.icon}

    def init_state(self):
        return None



# coding: utf-8

from supervisely.collection.key_indexed_collection import KeyIndexedCollection
from supervisely.io.json import JsonSerializable
from supervisely.annotation.tag_meta import TagMeta


class TagMetaCollection(KeyIndexedCollection, JsonSerializable):
    """
    Collection that stores TagMeta instances with unique names.
    """
    item_type = TagMeta

    def to_json(self):
        """
        Converts collection to json serializable list.
        Returns: json serializable dictionary
        """
        return [tag_meta.to_json() for tag_meta in self]

    @classmethod
    def from_json(cls, data):
        """
        Creates collection from json serializable list.
        Returns: TagMetaCollection
        """
        tags = [TagMeta.from_json(tag_meta_json) for tag_meta_json in data]
        return cls(tags)

    def get_id_mapping(self, raise_if_no_id=False):
        res = {}
        without_id = []
        for tag_meta in self:
            if tag_meta.sly_id is not None:
                if tag_meta.sly_id in res:
                    raise KeyError(f"TagMeta with id={tag_meta.sly_id} already exists (duplication). "
                                   f"Please contact tech support")
                else:
                    res[tag_meta.sly_id] = tag_meta
            else:
                without_id.append(tag_meta)
        if len(without_id) > 0 and raise_if_no_id is True:
            raise ValueError("There are TagMetas without id")
        return res


def make_renamed_tag_metas(src_tag_metas: TagMetaCollection, renamer, skip_missing=False) -> TagMetaCollection:
    '''
    The function make_renamed_classes rename classes names in given collection and return new collection
    :return: TagMetaCollection
    '''
    result_tags = []
    for src_tag in src_tag_metas:
        renamed_name = renamer.rename(src_tag.name)
        if renamed_name is not None:
            result_tags.append(src_tag.clone(name=renamed_name))
        elif not skip_missing:
            raise KeyError(
                'Tag meta named {} could not be mapped to a destination name.'.format(src_tag.name))
    return TagMetaCollection(items=result_tags)




# coding: utf-8

from copy import deepcopy

from supervisely.geometry.constants import X, Y, Z, \
    POSITION, ROTATION, DIMENTIONS, LABELER_LOGIN, UPDATED_AT, CREATED_AT, ID, CLASS_ID
from supervisely.geometry.geometry import Geometry


class Vector3d:
    '''
    This is a class for creating and using Vector3d objects for Cuboid3d class objects
    '''
    def __init__(self, x, y, z):
        '''
        :param x: int
        :param y: int
        :param z: int
        '''
        self._x = x
        self._y = y
        self._z = z

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    @property
    def z(self):
        return self._z

    def to_json(self):
        '''
        The function to_json convert Vector3d class object to json format(dict)
        :return: Vector3d in json format
        '''
        return {X: self.x, Y: self.y, Z: self.z}

    @classmethod
    def from_json(cls, data):
        '''
        The function from_json convert Vector3d from json format(dict) to Vector3d class object.
        :param data: Vector3d in json format(dict)
        :return: Vector3d class object
        '''
        x = data[X]
        y = data[Y]
        z = data[Z]
        return cls(x, y, z)

    def clone(self):
        return deepcopy(self)


class Cuboid3d(Geometry):
    '''
    This is a class for creating and using Cuboid3d objects for Labels
    '''
    @staticmethod
    def geometry_name():
        return 'cuboid_3d'

    def __init__(self, position: Vector3d, rotation: Vector3d, dimensions: Vector3d,
                 sly_id=None, class_id=None, labeler_login=None, updated_at=None, created_at=None):

        '''

        :param position: Vector3d class object
        :param rotation: Vector3d class object
        :param dimensions: Vector3d class object
        '''         
        super().__init__(sly_id=sly_id, class_id=class_id, labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)
        
        if type(position) is not Vector3d:
            raise TypeError("\"position\" param has to be of type {!r}".format(type(Vector3d)))
        if type(rotation) is not Vector3d:
            raise TypeError("\"rotation\" param has to be of type {!r}".format(type(Vector3d)))
        if type(dimensions) is not Vector3d:
            raise TypeError("\"dimensions\" param has to be of type {!r}".format(type(Vector3d)))

        self._position = position
        self._rotation = rotation
        self._dimensions = dimensions

    @property
    def position(self):
        return self._position.clone()

    @property
    def rotation(self):
        return self._rotation.clone()

    @property
    def dimensions(self):
        return self._dimensions.clone()

    def to_json(self):
        '''
        The function to_json convert Cuboid3d class object to json format(dict)
        :return: Cuboid3d in json format
        '''
        res = {POSITION: self.position.to_json(),
                ROTATION: self.rotation.to_json(),
                DIMENTIONS: self.dimensions.to_json()}

        self._add_creation_info(res)
        return res

    @classmethod
    def from_json(cls, data):
        '''
        The function from_json convert Cuboid3d from json format(dict) to Cuboid3d class object.
        :param data: Cuboid3d in json format(dict)
        :return: Cuboid3d class object
        '''
        position = Vector3d.from_json(data[POSITION])
        rotation = Vector3d.from_json(data[ROTATION])
        dimentions = Vector3d.from_json(data[DIMENTIONS])

        labeler_login = data.get(LABELER_LOGIN, None)
        updated_at = data.get(UPDATED_AT, None)
        created_at = data.get(CREATED_AT, None)
        sly_id = data.get(ID, None)
        class_id = data.get(CLASS_ID, None)
        return cls(position, rotation, dimentions, sly_id=sly_id, class_id=class_id, labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)



from supervisely.app.widgets.widget import Widget
from supervisely.app.widgets.radio_table.radio_table import RadioTable
from supervisely.app.widgets.notification_box.notification_box import NotificationBox
from supervisely.app.widgets.done_label.done_label import DoneLabel

# from supervisely.app.widgets.restart_step.restart_step import RestartStep



# coding: utf-8

from copy import deepcopy
import uuid

from supervisely._utils import take_with_default
from supervisely.video_annotation.video_tag_collection import VideoTagCollection
from supervisely.video_annotation.video_object_collection import VideoObjectCollection
from supervisely.video_annotation.frame_collection import FrameCollection
from supervisely.video_annotation.constants import FRAMES, IMG_SIZE, IMG_SIZE_HEIGHT, IMG_SIZE_WIDTH, \
                                                       DESCRIPTION, FRAMES_COUNT, TAGS, OBJECTS, VIDEO_ID, KEY, \
                                                       VIDEOS_MAP, VIDEO_NAME
from supervisely.video_annotation.key_id_map import KeyIdMap

from supervisely.video_annotation.video_annotation import VideoAnnotation
from supervisely.video_annotation.constants import FIGURES
from supervisely.pointcloud_annotation.constants import POINTCLOUD_ID
from supervisely.pointcloud_annotation.pointcloud_figure import PointcloudFigure
from supervisely.pointcloud_annotation.pointcloud_object_collection import  PointcloudObjectCollection


class PointcloudAnnotation(VideoAnnotation):
    '''
    This is a class for creating and using PointcloudAnnotation
    '''
    def __init__(self, objects=None, figures=None, tags=None, description="", key=None):
        '''
        :param objects: VideoObjectCollection
        :param figures: list of figures(Point, Cuboid, etc)
        :param tags: VideoTagCollection
        :param description: str
        :param key: uuid class object
        '''
        self._description = description
        self._tags = take_with_default(tags, VideoTagCollection())
        self._objects = take_with_default(objects, VideoObjectCollection())
        self._figures = take_with_default(figures, [])
        self._key = take_with_default(key, uuid.uuid4())

    @property
    def img_size(self):
        raise RuntimeError("Not supported for pointcloud")

    @property
    def frames_count(self):
        raise RuntimeError("Not supported for pointcloud")

    @property
    def figures(self):
        return deepcopy(self._figures)

    def validate_figures_bounds(self):
        raise RuntimeError("Not supported for pointcloud")

    def to_json(self, key_id_map: KeyIdMap=None):
        '''
        The function to_json convert PointcloudAnnotation to json format
        :param key_id_map: KeyIdMap class object
        :return: PointcloudAnnotation in json format
        '''
        res_json = {
            DESCRIPTION: self.description,
            KEY: self.key().hex,
            TAGS: self.tags.to_json(key_id_map),
            OBJECTS: self.objects.to_json(key_id_map),
            FIGURES: [figure.to_json(key_id_map) for figure in self.figures]
        }

        if key_id_map is not None:
            pointcloud_id = key_id_map.get_video_id(self.key())
            if pointcloud_id is not None:
                res_json[POINTCLOUD_ID] = pointcloud_id

        return res_json

    @classmethod
    def from_json(cls, data, project_meta, key_id_map: KeyIdMap=None):
        '''
        :param data: input PointcloudAnnotation in json format
        :param project_meta: ProjectMeta class object
        :param key_id_map: KeyIdMap class object
        :return: PointcloudAnnotation class object
        '''
        item_key = uuid.UUID(data[KEY]) if KEY in data else uuid.uuid4()
        if key_id_map is not None:
            key_id_map.add_video(item_key, data.get(POINTCLOUD_ID, None))
        description = data.get(DESCRIPTION, "")
        tags = VideoTagCollection.from_json(data[TAGS], project_meta.tag_metas, key_id_map)
        objects = PointcloudObjectCollection.from_json(data[OBJECTS], project_meta, key_id_map)

        figures = []
        for figure_json in data.get(FIGURES, []):
            figure = PointcloudFigure.from_json(figure_json, objects, None, key_id_map)
            figures.append(figure)

        return cls(objects=objects,
                   figures=figures,
                   tags=tags,
                   description=description,
                   key=item_key)

    def clone(self, objects=None, figures=None, tags=None, description=None):
        '''
        :param objects: VideoObjectCollection
        :param figures: list of figures
        :param tags: VideoTagCollection
        :param description: str
        :return: PointcloudAnnotation class object
        '''
        return PointcloudAnnotation(objects=take_with_default(objects, self.objects),
                                    figures=take_with_default(figures, self.figures),
                                    tags=take_with_default(tags, self.tags),
                                    description=take_with_default(description, self.description))



# coding: utf-8

import uuid
from supervisely.annotation.tag import Tag, TagJsonFields
from supervisely._utils import take_with_default
from supervisely.video_annotation.constants import KEY, ID, FRAME_RANGE
from supervisely.video_annotation.key_id_map import KeyIdMap


class VideoTag(Tag):
    '''
    This is a class for creating and using VideoTag objects for videos
    '''
    def __init__(self, meta, value=None, frame_range=None, key=None, sly_id=None, labeler_login=None, updated_at=None, created_at=None):
        '''
        :param meta: Tag metadata: it include tag name, value type, and possible values for tags with enum values.
        When creating a new tag, the value is automatically cross-checked against the metadata to make sure the value
        is valid.
        :param value: There are 3 possible value types of value: ANY_NUMBER for numeric values,
        ANY_STRING for arbitrary string values, ONEOF_STRING for string values restricted to a given whitelist
        :param frame_range: tuple or list of integers
        :param key: uuid class object
        '''
        super(VideoTag, self).__init__(meta, value=value, sly_id=sly_id, labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)
        
        self._frame_range = None
        if frame_range is not None:
            if not isinstance(frame_range, (tuple, list)):
                raise TypeError('{!r} has to be a tuple or a list. Given type "{}".'.format(FRAME_RANGE, type(frame_range)))
            else:
                self._frame_range = list(frame_range)

        self._key = take_with_default(key, uuid.uuid4())

    @property
    def frame_range(self):
        return self._frame_range

    def key(self):
        return self._key

    def to_json(self, key_id_map: KeyIdMap = None):
        '''
        The function to_json convert VideoTag class object to json format
        :param key_id_map: KeyIdMap class object
        :return: VideoTag in json format
        '''
        data_json = super(VideoTag, self).to_json()
        if type(data_json) is str:
            # @TODO: case when tag has no value, super.to_json() returns tag name
            data_json = {TagJsonFields.TAG_NAME: data_json}
        if self.frame_range is not None:
            data_json[FRAME_RANGE] = self.frame_range
        data_json[KEY] = self.key().hex

        if key_id_map is not None:
            item_id = key_id_map.get_tag_id(self.key())
            if item_id is not None:
                data_json[ID] = item_id

        return data_json

    @classmethod
    def from_json(cls, data, tag_meta_collection, key_id_map: KeyIdMap = None):
        '''
        The function from_json convert VideoTag from json format to VideoTag class object.
        :param data: input VideoTag in json format
        :param tag_meta_collection: VideoTagCollection
        :param key_id_map: KeyIdMap class object
        :return: VideoTag class object
        '''
        temp = super(VideoTag, cls).from_json(data, tag_meta_collection)
        frame_range = data.get(FRAME_RANGE, None)
        key = uuid.UUID(data[KEY]) if KEY in data else uuid.uuid4()

        if key_id_map is not None:
            key_id_map.add_tag(key, data.get(ID, None))

        return cls(meta=temp.meta, value=temp.value, frame_range=frame_range, key=key,
                   sly_id=temp.sly_id, labeler_login=temp.labeler_login, updated_at=temp.updated_at, created_at=temp.created_at)

    def get_compact_str(self):
        '''
        :return: string with information about tag(name, value) and range of frames
        '''
        res = super(VideoTag, self).get_compact_str()
        if self.frame_range is not None:
            res = "{}[{} - {}]".format(res, self.frame_range[0], self.frame_range[1])
        return res

    def __eq__(self, other):
        return isinstance(other, VideoTag) and \
               self.meta == other.meta and \
               self.value == other.value and \
               self.frame_range == other.frame_range

    def clone(self, meta=None, value=None, frame_range=None, key=None,
                    sly_id=None, labeler_login=None, updated_at=None, created_at=None):
        '''
        :param meta: Tag metadata
        :param value: There are 3 possible value types of value: ANY_NUMBER for numeric values,
        ANY_STRING for arbitrary string values, ONEOF_STRING for string values restricted to a given whitelist
        :param frame_range: tuple or list of integers
        :param key: uuid class object
        :return: VideoTag class object
        '''
        return VideoTag(meta=take_with_default(meta, self.meta),
                        value=take_with_default(value, self.value),
                        frame_range=take_with_default(frame_range, self.frame_range),
                        key=take_with_default(key, self.key),
                        sly_id=take_with_default(sly_id, self.sly_id),
                        labeler_login=take_with_default(labeler_login, self.labeler_login),
                        updated_at=take_with_default(updated_at, self.updated_at),
                        created_at=take_with_default(created_at, self.created_at))

    def __str__(self):
        return '{:<7s}{:<10}{:<7s} {:<13}{:<7s} {:<10} {:<12}'.format('Name:', self._meta.name,
                                                               'Value type:', self._meta.value_type,
                                                               'Value:', str(self.value),
                                                               'FrameRange', str(self.frame_range))

    @classmethod
    def get_header_ptable(cls):
        return ['Name', 'Value type', 'Value', 'Frame range']

    def get_row_ptable(self):
        return [self._meta.name, self._meta.value_type, self.value, self.frame_range]



# coding: utf-8
from __future__ import annotations

import json
import itertools
import numpy as np
from typing import List
import operator
import cv2
from copy import deepcopy
from PIL import Image

from supervisely import logger
from supervisely.annotation.label import Label
from supervisely.annotation.obj_class_collection import ObjClassCollection
from supervisely.annotation.tag_collection import TagCollection
from supervisely.project.project_meta import ProjectMeta
from supervisely.geometry.rectangle import Rectangle
from supervisely.imaging import image as sly_image
from supervisely.imaging import font as sly_font
from supervisely._utils import take_with_default
from supervisely.geometry.multichannel_bitmap import MultichannelBitmap
from supervisely.geometry.bitmap import Bitmap
from supervisely.geometry.polygon import Polygon
from supervisely.io.fs import ensure_base_path


ANN_EXT = '.json'


class AnnotationJsonFields:
    IMG_DESCRIPTION = 'description'
    IMG_SIZE = 'size'
    IMG_SIZE_WIDTH = 'width'
    IMG_SIZE_HEIGHT = 'height'
    IMG_TAGS = 'tags'
    LABELS = 'objects'
    CUSTOM_DATA = "customBigData"
    PROBABILITY_CLASSES = "probabilityClasses"
    PROBABILITY_LABELS = "probabilityLabels"
    IMAGE_ID = "imageId"


class Annotation:
    '''
    This is a class for creating and using annotations for images

    Attributes:
        img_size (tuple): size of the image
        labels (list): list of Label class objects
        img_tags (list): list of image tags
        img_description (str): image description
        pixelwise_scores_labels (list)
    '''
    def __init__(self, img_size, labels=None, img_tags=None, img_description="",
                 pixelwise_scores_labels=None, custom_data=None, image_id=None):
        '''
        The constructor for Annotation class.
        :param img_size(tuple): size of the image
        :param labels(list): list of Label class objects
        :param img_tags(list): list of image tags
        :param img_description(str): image description
        :param pixelwise_scores_labels(list)
        '''
        if not isinstance(img_size, (tuple, list)):
            raise TypeError('{!r} has to be a tuple or a list. Given type "{}".'.format('img_size', type(img_size)))
        self._img_size = tuple(img_size)
        self._img_description = img_description
        self._img_tags = take_with_default(img_tags, TagCollection())
        self._labels = []
        self._add_labels_impl(self._labels, take_with_default(labels, []))
        self._pixelwise_scores_labels = []  # @TODO: store pixelwise scores as usual geometry labels
        self._add_labels_impl(self._pixelwise_scores_labels, take_with_default(pixelwise_scores_labels, []))
        self._custom_data = take_with_default(custom_data, {})
        self._image_id = image_id

    @property
    def img_size(self):
        return deepcopy(self._img_size)

    @property
    def image_id(self):
        return self._image_id

    @property
    def labels(self) -> List[Label]:
        return self._labels.copy()

    @property
    def pixelwise_scores_labels(self):
        return self._pixelwise_scores_labels.copy()

    @property
    def img_description(self):
        return self._img_description

    @property
    def img_tags(self) -> TagCollection:
        return self._img_tags

    def to_json(self):
        '''
        The function to_json convert annotation to json format
        :return: annotation in json format
        '''
        res = {
            AnnotationJsonFields.IMG_DESCRIPTION: self.img_description,
            AnnotationJsonFields.IMG_SIZE: {
                AnnotationJsonFields.IMG_SIZE_HEIGHT: int(self.img_size[0]),
                AnnotationJsonFields.IMG_SIZE_WIDTH: int(self.img_size[1])
            },
            AnnotationJsonFields.IMG_TAGS: self.img_tags.to_json(),
            AnnotationJsonFields.LABELS: [label.to_json() for label in self.labels],
            AnnotationJsonFields.CUSTOM_DATA: self.custom_data
        }
        if len(self._pixelwise_scores_labels) > 0:
            # construct probability classes from labels
            prob_classes = {}
            for label in self._pixelwise_scores_labels:
                # @TODO: hotfix to save geometry as "multichannelBitmap" instead of "bitmap"; use normal classes
                prob_classes[label.obj_class.name] = label.obj_class.clone(geometry_type=MultichannelBitmap)

            # save probabilities
            probabilities = {
                AnnotationJsonFields.PROBABILITY_LABELS: [label.to_json() for label in self._pixelwise_scores_labels],
                AnnotationJsonFields.PROBABILITY_CLASSES: ObjClassCollection(list(prob_classes.values())).to_json()
            }
            res[AnnotationJsonFields.CUSTOM_DATA].update(probabilities)
        if self.image_id is not None:
            res[AnnotationJsonFields.IMAGE_ID] = self.image_id
        return res

    @classmethod
    def from_json(cls, data, project_meta):
        '''
        The function from_json convert annotation from json format to Annotation class object. If one of the labels
        of annotation in json format cannot be convert to Label class object it generate exception error.
        :param data: input annotation in json format
        :param project_meta: ProjectMeta class object
        :return: Annotation class object
        '''
        img_size_dict = data[AnnotationJsonFields.IMG_SIZE]
        img_height = img_size_dict[AnnotationJsonFields.IMG_SIZE_HEIGHT]
        img_width = img_size_dict[AnnotationJsonFields.IMG_SIZE_WIDTH]
        img_size = (img_height, img_width)
        try:
            labels = [Label.from_json(label_json, project_meta) for label_json in data[AnnotationJsonFields.LABELS]]
        except Exception:
            logger.fatal('Failed to deserialize annotation from JSON format. One of the Label objects could not be '
                         'deserialized')
            raise

        custom_data = data.get(AnnotationJsonFields.CUSTOM_DATA, {})
        prob_labels = None
        if AnnotationJsonFields.PROBABILITY_LABELS in custom_data and \
                AnnotationJsonFields.PROBABILITY_CLASSES in custom_data:

            prob_classes = ObjClassCollection.from_json(custom_data[AnnotationJsonFields.PROBABILITY_CLASSES])

            # @TODO: tony, maybe link with project meta (add probability classes???)
            prob_project_meta = ProjectMeta(obj_classes=prob_classes)
            prob_labels = [Label.from_json(label_json, prob_project_meta)
                           for label_json in custom_data[AnnotationJsonFields.PROBABILITY_LABELS]]

            custom_data.pop(AnnotationJsonFields.PROBABILITY_CLASSES)
            custom_data.pop(AnnotationJsonFields.PROBABILITY_LABELS)

        image_id = data.get(AnnotationJsonFields.IMAGE_ID, None)

        return cls(img_size=img_size,
                   labels=labels,
                   img_tags=TagCollection.from_json(data[AnnotationJsonFields.IMG_TAGS], project_meta.tag_metas),
                   img_description=data.get(AnnotationJsonFields.IMG_DESCRIPTION, ""),
                   pixelwise_scores_labels=prob_labels,
                   custom_data=custom_data,
                   image_id=image_id)

    @classmethod
    def load_json_file(cls, path, project_meta):
        '''
        The function load_json_file download json file and convert in to the Annotation class object
        :param path: the path to the input json file
        :param project_meta: ProjectMeta class object
        :return: Annotation class object
        '''
        with open(path) as fin:
            data = json.load(fin)
        return cls.from_json(data, project_meta)

    def clone(self, img_size=None, labels=None, img_tags=None, img_description=None,
              pixelwise_scores_labels=None, custom_data=None, image_id=None):
        '''
        The function clone make copy of the Annotation class object
        :return: Annotation class object
        '''
        return Annotation(img_size=take_with_default(img_size, self.img_size),
                          labels=take_with_default(labels, self.labels),
                          img_tags=take_with_default(img_tags, self.img_tags),
                          img_description=take_with_default(img_description, self.img_description),
                          pixelwise_scores_labels=take_with_default(pixelwise_scores_labels, self.pixelwise_scores_labels),
                          custom_data=take_with_default(custom_data, self.custom_data),
                          image_id = take_with_default(image_id, self.image_id)
                          )

    def _add_labels_impl(self, dest, labels):
        '''
        The function _add_labels_impl extend list of the labels of the current Annotation object
        :param dest: destination list of the Label class objects
        :param labels: list of the Label class objects to be added to the destination list
        :return: list of the Label class objects
        '''
        for label in labels:
            # TODO Reconsider silent automatic normalization, reimplement
            canvas_rect = Rectangle.from_size(self.img_size)
            dest.extend(label.crop(canvas_rect))

    def add_label(self, label):
        '''
        The function add_label add label to the current Annotation object and return the copy of the
        current  Annotation object
        :param label: Label class object to be added
        :return: Annotation class object with the new list of the Label class objects
        '''
        return self.add_labels([label])

    def add_labels(self, labels):
        '''
        The function add_labels extend list of the labels of the current Annotation object and return the copy of the
        current  Annotation object
        :param labels: list of the Label class objects to be added
        :return: Annotation class object with the new list of the Label class objects
        '''
        new_labels = []
        self._add_labels_impl(new_labels, labels)
        return self.clone(labels=[*self._labels, *new_labels])

    def delete_label(self, label):
        '''
        The function delete_label detele label from the current Annotation object and return the copy of the
        current  Annotation object. If there is no deleted label in current Annotation object it generate exception
        error(KeyError).
        :param label: Label class object to be delete
        :return: Annotation class object with the new list of the Label class objects
        '''
        retained_labels = [_label for _label in self._labels if _label != label]
        if len(retained_labels) == len(self._labels):
            raise KeyError('Trying to delete a non-existing label of class: {}'.format(label.obj_class.name))
        return self.clone(labels=retained_labels)

    def add_pixelwise_score_label(self, label):
        '''
        The function add_pixelwise_score_label add label to the pixelwise_scores_labels and return the copy of the
        current  Annotation object
        :param label: Label class object to be added
        :return: Annotation class object with the new list of the pixelwise_scores_labels
        '''
        return self.add_pixelwise_score_labels([label])

    def add_pixelwise_score_labels(self, labels):
        '''
        The function add_pixelwise_score_labels extend list of the labels of the pixelwise_scores_labels and return
        the copy of the current  Annotation object.
        :param labels: list of the Label class objects to be added
        :return: Annotation class object with the new list of the pixelwise_scores_labels
        '''
        new_labels = []
        self._add_labels_impl(new_labels, labels)
        return self.clone(pixelwise_scores_labels=[*self._pixelwise_scores_labels, *new_labels])

    def add_tag(self, tag):
        '''
        The function add_tag add tag to the current Annotation object and return the copy of the
        current  Annotation object
        :param tag: TagCollection class object to be added
        :return: Annotation class object with the new list of the tags
        '''
        return self.clone(img_tags=self._img_tags.add(tag))

    def add_tags(self, tags):
        '''
        The function add_tags add tags to the current Annotation object and return the copy of the
        current  Annotation object
        :param tags: list of the TagCollection class objects to be added
        :return: Annotation class object with the new list of the tags
        '''
        return self.clone(img_tags=self._img_tags.add_items(tags))

    def delete_tags_by_name(self, tag_names):
        '''
        The function delete_tags_by_name removes tags by their names from current Annotation object and return the copy
        of the current  Annotation object
        :param tag_names: list of the tag names to be delete
        :return: Annotation class object with the new list of the tags
        '''
        retained_tags = [tag for tag in self._img_tags.items() if tag.meta.name not in tag_names]
        return self.clone(img_tags=TagCollection(items=retained_tags))

    def delete_tag_by_name(self, tag_name):
        '''
        The function delete_tag_by_name removes tag by it name from current Annotation object and return the copy
        of the current  Annotation object
        :param tag_name: tag names to be delete
        :return: Annotation class object with the new list of the tags
        '''
        return self.delete_tags_by_name([tag_name])

    def delete_tags(self, tags):
        '''
        The function delete_tags removes tags from current Annotation object and return the copy of the current
        Annotation object
        :param tags: list of the TagCollection class objects to be deleted
        :return: Annotation class object with the new list of the tags
        '''
        return self.delete_tags_by_name({tag.meta.name for tag in tags})

    def delete_tag(self, tag):
        '''
        The function delete_tag remove tag from current Annotation object and return the copy of the current
        Annotation object
        :param tag: TagCollection class object to be deleted
        :return: Annotation class object with the new list of the tags
        '''
        return self.delete_tags_by_name([tag.meta.name])

    def transform_labels(self, label_transform_fn, new_size=None):
        '''
        The function transform_labels transform labels and change image size in current Annotation object and return the copy of the current
        Annotation object
        :param label_transform_fn: function for transform labels
        :param new_size: new image size
        :return: Annotation class object with new labels and image size
        '''
        def _do_transform_labels(src_labels, label_transform_fn):
            # long easy to debug
            # result = []
            # for label in src_labels:
            #     result.extend(label_transform_fn(label))
            # return result

            # short, hard-to-debug alternative
            return list(itertools.chain(*[label_transform_fn(label) for label in src_labels]))
        new_labels = _do_transform_labels(self._labels, label_transform_fn)
        new_pixelwise_scores_labels = _do_transform_labels(self._pixelwise_scores_labels, label_transform_fn)
        return self.clone(img_size=take_with_default(new_size, self.img_size), labels=new_labels,
                          pixelwise_scores_labels=new_pixelwise_scores_labels)

    def crop_labels(self, rect):
        '''
        The function crop_labels crops labels in current Annotation object and return the copy of the current
        Annotation object
        :param rect: Rectangle class object
        :return: Annotation class object with new labels
        '''
        def _crop_label(label):
            return label.crop(rect)
        return self.transform_labels(_crop_label)

    def relative_crop(self, rect):
        '''
        The function relative_crop crops labels and change image size in current Annotation object and return the copy of the current
        Annotation object
        :param rect: Rectangle class object
        :return: Annotation class object with new labels and image size
        '''
        def _crop_label(label):
            return label.relative_crop(rect)
        return self.transform_labels(_crop_label, rect.to_size())

    def rotate(self, rotator):
        '''
        The function rotates all labels of the current Annotation object and return the copy of the current
        Annotation object
        :param rotator: ImageRotator class object
        :return: Annotation class object with new(rotated) labels and image size
        '''
        def _rotate_label(label):
            return [label.rotate(rotator)]
        return self.transform_labels(_rotate_label, tuple(rotator.new_imsize))

    def resize(self, out_size):
        '''
        The function resize all labels of the current Annotation object and return the copy of the current
        Annotation object
        :param out_size: new image size
        :return: Annotation class object with new(resized) labels and image size
        '''
        def _resize_label(label):
            return [label.resize(self.img_size, out_size)]
        return self.transform_labels(_resize_label, out_size)

    def scale(self, factor):
        '''
        The function scale change scale of the current Annotation object with a given factor and return the copy of the
        current Annotation object
        :param factor: float scale parameter
        :return: Annotation class object with new scale
        '''
        def _scale_label(label):
            return [label.scale(factor)]
        result_size = (round(self.img_size[0] * factor), round(self.img_size[1] * factor))
        return self.transform_labels(_scale_label, result_size)

    def fliplr(self):
        '''
        The function flip the current Annotation object in horizontal and return the copy of the
        current Annotation object
        :return: Annotation class object
        '''
        def _fliplr_label(label):
            return [label.fliplr(self.img_size)]
        return self.transform_labels(_fliplr_label)

    def flipud(self):
        '''
        The function flip the current Annotation object in vertical and return the copy of the
        current Annotation object
        :return: Annotation class object
        '''
        def _flipud_label(label):
            return [label.flipud(self.img_size)]
        return self.transform_labels(_flipud_label)

    def _get_font(self):
        '''
        The function get size of font for image with given size
        :return: font for drawing
        '''
        return sly_font.get_font(font_size=sly_font.get_readable_font_size(self.img_size))

    def _draw_tags(self, bitmap):
        '''
        The function draws text labels on bitmap from left to right.
        :param bitmap: target image
        '''
        texts = [tag.get_compact_str() for tag in self.img_tags]
        sly_image.draw_text_sequence(bitmap, texts, (0, 0), sly_image.CornerAnchorMode.TOP_LEFT,
                                     font=self._get_font())

    def draw(self, bitmap, color=None, thickness=1, draw_tags=False):
        '''
        The function draws rectangle near each label on bitmap
        :param bitmap: target image (np.ndarray)
        :param color: [R, G, B]
        :param thickness: thickness of the drawing rectangle
        '''
        tags_font = None
        if draw_tags is True:
            tags_font = self._get_font()
        for label in self._labels:
            label.draw(bitmap, color=color, thickness=thickness, draw_tags=draw_tags, tags_font=tags_font)
        if draw_tags:
            self._draw_tags(bitmap)

    def draw_contour(self, bitmap, color=None, thickness=1, draw_tags=False):
        '''
        The function draws the figure contour on a given bitmap
        :param bitmap: target image (np.ndarray)
        :param color: [R, G, B]
        :param thickness: thickness of the drawing contour
        '''
        tags_font = None
        if draw_tags is True:
            tags_font = self._get_font()
        for label in self._labels:
            label.draw_contour(bitmap, color=color, thickness=thickness, draw_tags=draw_tags, tags_font=tags_font)
        if draw_tags:
            self._draw_tags(bitmap)

    @classmethod
    def from_img_path(cls, img_path):
        '''
        The function from_img_path download image on the given path and return size of the image
        :param img_path: the path to the input image
        :return: size of the image
        '''
        img = sly_image.read(img_path)
        img_size = img.shape[:2]
        return cls(img_size)

    @classmethod
    def stat_area(cls, render, names, colors):
        if len(names) != len(colors):
            raise RuntimeError("len(names) != len(colors) [{} != {}]".format(len(names), len(colors)))

        result = {}

        height, width = render.shape[:2]
        total_pixels = height * width

        channels = None
        if len(render.shape) == 2:
            channels = 1
        elif len(render.shape) == 3:
            channels = render.shape[2]

        unlabeled_done = False

        covered_pixels = 0
        for name, color in zip(names, colors):
            col_name = name
            if name == "unlabeled":
                unlabeled_done = True
            class_mask = np.all(render == color, axis=-1).astype('uint8')
            cnt_pixels = class_mask.sum()
            covered_pixels += cnt_pixels
            result[col_name] = cnt_pixels / total_pixels * 100.0

        if covered_pixels > total_pixels:
            raise RuntimeError("Class colors mistake: covered_pixels > total_pixels")

        if unlabeled_done is False:
            result['unlabeled'] = (total_pixels - covered_pixels) / total_pixels * 100.0

        result['height'] = height
        result['width'] = width
        result['channels'] = channels
        return result

    def stat_class_count(self, class_names):
        total = 0
        stat = {name: 0 for name in class_names}
        for label in self._labels:
            cur_name = label.obj_class.name
            if cur_name not in stat:
                raise KeyError("Class {!r} not found in {}".format(cur_name, class_names))
            stat[cur_name] += 1
            total += 1
        stat['total'] = total
        return stat

    def draw_class_idx_rgb(self, render, name_to_index):
        for label in self._labels:
            class_idx = name_to_index[label.obj_class.name]
            color = [class_idx, class_idx, class_idx]
            label.draw(render, color=color, thickness=1)

    @property
    def custom_data(self):
        return self._custom_data.copy()

    def filter_labels_by_min_side(self, thresh, filter_operator=operator.lt, classes=None):
        def filter(label):
            if classes == None or label.obj_class.name in classes:
                bbox = label.geometry.to_bbox()
                height_px = bbox.height
                width_px = bbox.width
                if filter_operator(min(height_px, width_px), thresh):
                    return []  # action 'delete'
            return [label]
        return self.transform_labels(filter)

    def get_label_by_id(self, sly_id) -> Label:
        for label in self._labels:
            if label.geometry.sly_id == sly_id:
                return label
        return None

    def merge(self, other: Annotation):
        res = self.clone()
        res = res.add_labels(other.labels)
        res = res.add_tags(other.img_tags)
        return res

    def draw_pretty(self, bitmap: np.ndarray, color: List[int, int, int] = None, thickness: int = 1,
                    opacity: float = 0.5, draw_tags: bool = False, output_path: str = None) -> None:
        height, width = bitmap.shape[:2]
        vis_filled = np.zeros((height, width, 3), np.uint8)
        self.draw(vis_filled, color=color, thickness=thickness, draw_tags=draw_tags)
        vis = cv2.addWeighted(bitmap, 1, vis_filled, opacity, 0)
        np.copyto(bitmap, vis)
        if thickness > 0:
            self.draw_contour(bitmap, color=color, thickness=thickness, draw_tags=draw_tags)
        if output_path:
            sly_image.write(output_path, bitmap)

    def to_nonoverlapping_masks(self, mapping):
        common_img = np.zeros(self.img_size, np.int32)  # size is (h, w)
        for idx, lbl in enumerate(self.labels, start=1):
            #if mapping[lbl.obj_class] is not None:
            lbl.draw(common_img, color=idx)
        #(unique, counts) = np.unique(common_img, return_counts=True)
        new_labels = []
        for idx, lbl in enumerate(self.labels, start=1):

            dest_class = mapping[lbl.obj_class]
            if dest_class is None:
                continue  # skip labels

            mask = common_img == idx
            if np.any(mask):  # figure may be entirely covered by others
                g = lbl.geometry
                new_mask = Bitmap(data=mask)
                new_lbl = lbl.clone(geometry=new_mask, obj_class=dest_class)
                new_labels.append(new_lbl)
        new_ann = self.clone(labels=new_labels)
        return new_ann

    def to_indexed_color_mask(self, mask_path, palette=Image.ADAPTIVE, colors=256):
        mask = np.zeros((self.img_size[0], self.img_size[1], 3), dtype=np.uint8)
        for label in self.labels:
            label.geometry.draw(mask, label.obj_class.color)

        im = Image.fromarray(mask)
        im = im.convert("P", palette=palette, colors=colors)

        ensure_base_path(mask_path)
        im.save(mask_path)

    def to_segmentation_task(self):
        class_mask = {}
        for label in self.labels:
            if label.obj_class not in class_mask:
                class_mask[label.obj_class] = np.zeros(self.img_size, np.uint8)
            label.draw(class_mask[label.obj_class], color=255)
        new_labels = []
        for obj_class, white_mask in class_mask.items():
            mask = white_mask == 255
            bitmap = Bitmap(data=mask)
            new_labels.append(Label(geometry=bitmap, obj_class=obj_class))
        return self.clone(labels=new_labels)

    def to_detection_task(self, mapping):
        aux_mapping = mapping.copy()

        to_render_mapping = {}
        to_render_labels = []
        other_labels = []
        _polygons_to_bitmaps_classes = {}
        for lbl in self.labels:
            if type(lbl.geometry) in [Bitmap, Polygon]:
                to_render_labels.append(lbl)
                if type(lbl.geometry) is Polygon:
                    new_class = _polygons_to_bitmaps_classes.get(lbl.obj_class.name, None)
                    if new_class is None:
                        new_class = lbl.obj_class.clone(geometry_type=Bitmap)
                        _polygons_to_bitmaps_classes[lbl.obj_class.name] = new_class
                        aux_mapping[new_class] = aux_mapping[lbl.obj_class]
                    to_render_mapping[lbl.obj_class] = new_class

                else:
                    to_render_mapping[lbl.obj_class] = lbl.obj_class
            else:
                other_labels.append(lbl)
        ann_raster = self.clone(labels=to_render_labels)
        ann_raster = ann_raster.to_nonoverlapping_masks(to_render_mapping)
        new_labels = []
        for lbl in [*ann_raster.labels, *other_labels]:
            dest_class = aux_mapping[lbl.obj_class]
            if dest_class is None:
                continue  # skip labels
            if dest_class == lbl.obj_class:
                new_labels.append(lbl)
            else:
                bbox = lbl.geometry.to_bbox()
                new_lbl = lbl.clone(geometry=bbox, obj_class=dest_class)
                new_labels.append(new_lbl)
        new_ann = self.clone(labels=new_labels)
        return new_ann

    def masks_to_imgaug(self, class_to_index=None):
        from imgaug.augmentables.segmaps import SegmentationMapsOnImage
        h = self.img_size[0]
        w = self.img_size[1]
        mask = np.zeros((h, w, 1), dtype=np.int32)

        for index, label in enumerate(self.labels, start=1):
            label: Label
            if type(label.geometry) == Bitmap:
                if class_to_index is not None:
                    label.draw(mask, class_to_index[label.obj_class.name])
                else:
                    label.draw(mask, index)


        segmaps = None
        if np.any(mask):
            segmaps = SegmentationMapsOnImage(mask, shape=self.img_size)
        return segmaps

    def bboxes_to_imgaug(self):
        from imgaug.augmentables.bbs import BoundingBox, BoundingBoxesOnImage
        boxes = []
        for label in self.labels:
            if type(label.geometry) == Rectangle:
                rect: Rectangle = label.geometry
                boxes.append(
                    BoundingBox(x1=rect.left, y1=rect.top,
                                x2=rect.right, y2=rect.bottom,
                                label=label.obj_class.name)
                )
        bbs = None
        if len(boxes) > 0:
            bbs = BoundingBoxesOnImage(boxes, shape=self.img_size)
        return bbs

    @classmethod
    def from_imgaug(cls, img, ia_boxes=None, ia_masks=None,
                    index_to_class=None,
                    meta: ProjectMeta = None):
        if ((ia_boxes is not None) or (ia_masks is not None)) and meta is None:
            raise ValueError("Project meta has to be provided")

        labels = []
        if ia_boxes is not None:
            for ia_box in ia_boxes:
                obj_class = meta.get_obj_class(ia_box.label)
                if obj_class is None:
                    raise KeyError("Class {!r} not found in project meta".format(ia_box.label))
                lbl = Label(Rectangle(top=ia_box.y1, left=ia_box.x1, bottom=ia_box.y2, right=ia_box.x2), obj_class)
                labels.append(lbl)

        if ia_masks is not None:
            if index_to_class is None:
                raise ValueError("mapping from index to class name is needed to transform masks to SLY format")
            class_mask = ia_masks.get_arr()
            # mask = white_mask == 255
            (unique, counts) = np.unique(class_mask, return_counts=True)
            for index, count in zip(unique, counts):
                if index == 0:
                    continue
                mask = class_mask == index
                bitmap = Bitmap(data=mask[:, :, 0])
                restore_class = meta.get_obj_class(index_to_class[index])
                labels.append(Label(geometry=bitmap, obj_class=restore_class))

        return cls(img_size=img.shape[:2], labels=labels)

    def is_empty(self):
        if len(self.labels) == 0 and len(self.img_tags) == 0:
            return True
        else:
            return False

    def filter_labels_by_classes(self, keep_classes):
        new_labels = []
        for lbl in self.labels:
            if lbl.obj_class.name in keep_classes:
                new_labels.append(lbl.clone())
        return self.clone(labels=new_labels)



# coding: utf-8

from supervisely.video_annotation.video_object_collection import VideoObjectCollection
from supervisely.pointcloud_annotation.pointcloud_object import PointcloudObject


class PointcloudObjectCollection(VideoObjectCollection):
    '''
    Collection that stores PointcloudObject instances.
    '''
    item_type = PointcloudObject


# coding: utf-8
import json


class JsonSerializable:
    def to_json(self):
        """ Serialize to JSON-compatible dict.
        :return: dict
        """
        raise NotImplementedError()

    @classmethod
    def from_json(cls, data):
        """
        Deserialize from a JSON-compatible dict
        :param data: JSON-compatible dict
        :return: Parsed object
        """
        raise NotImplementedError()


def load_json_file(filename):
    '''
    Decoding data from json file with given filename
    :param filename: str
    :return: data in json format
    '''
    with open(filename, encoding='utf-8') as fin:
        return json.load(fin)


def dump_json_file(data, filename, indent=4):
    '''
    Write given data in json format in file with given name
    :param data: data to write in file in json format
    :param filename: str
    :param indent: int
    '''
    with open(filename, 'w') as fout:
        json.dump(data, fout, indent=indent)


def flatten_json(data, sep="."):
    import pandas as pd
    df = pd.json_normalize(data, sep=sep)
    return df.to_dict(orient='records')[0]


def modify_keys(data, prefix=None, suffix=None):
    def _modify(k):
        res = k
        if prefix is not None:
            res = prefix + res
        if suffix is not None:
            res += suffix
        return res
    return {_modify(k): v for k, v in data.items()}



from supervisely.app.widgets import Widget

INFO = "info"
WARNING = "warning"
ERROR = "error"


class DoneLabel(Widget):
    def __init__(
        self,
        text: str = None,
        widget_id: str = None,
    ):
        self.text = text
        super().__init__(widget_id=widget_id, file_path=__file__)

    def init_data(self):
        return {"text": self.text}

    def init_state(self):
        return None



# coding: utf-8

import uuid
from bidict import bidict

from supervisely.api.module_api import ApiField

from supervisely._utils import take_with_default
from supervisely.geometry.any_geometry import AnyGeometry
from supervisely.geometry.rectangle import Rectangle
from supervisely.video_annotation.video_object_collection import VideoObjectCollection
from supervisely.video_annotation.constants import ID, KEY, OBJECT_ID, OBJECT_KEY, META
from supervisely.api.module_api import ApiField
from supervisely.annotation.json_geometries_map import GET_GEOMETRY_FROM_STR
from supervisely.video_annotation.key_id_map import KeyIdMap

from supervisely.geometry.constants import LABELER_LOGIN, UPDATED_AT, CREATED_AT, CLASS_ID


class OutOfImageBoundsExtension(Exception):
    pass


class VideoFigure:
    def __init__(self, video_object, geometry, frame_index, key=None, class_id=None, labeler_login=None, updated_at=None, created_at=None):
        self._video_object = video_object
        self._set_geometry_inplace(geometry)
        self._frame_index = frame_index
        self._key = take_with_default(key, uuid.uuid4())
        self.class_id = class_id
        self.labeler_login = labeler_login
        self.updated_at = updated_at
        self.created_at = created_at

    def _add_creation_info(self, d):
        if self.labeler_login is not None:
            d[LABELER_LOGIN] = self.labeler_login
        if self.updated_at is not None:
            d[UPDATED_AT] = self.updated_at
        if self.created_at is not None:
            d[CREATED_AT] = self.created_at

    def _set_geometry_inplace(self, geometry):
        '''
        Checks the given geometry for correctness. Raise error if given geometry type != geometry type of VideoObject class
        :param geometry: Geometry class object (Point, Rectangle etc)
        '''
        self._geometry = geometry
        self._validate_geometry_type()
        self._validate_geometry()

    @property
    def video_object(self):
        return self._video_object

    @property
    def parent_object(self):
        return self._video_object

    @property
    def geometry(self):
        return self._geometry

    @property
    def frame_index(self):
        return self._frame_index

    def key(self):
        return self._key

    def _validate_geometry(self):
        '''
        Checks geometry of VideoFigure class object for correctness
        '''
        self._geometry.validate(self.parent_object.obj_class.geometry_type.geometry_name(),
                                self.parent_object.obj_class.geometry_config)

    def _validate_geometry_type(self):
        '''
        Raise error if given geometry type != geometry type of VideoObject class
        '''
        if self.parent_object.obj_class.geometry_type != AnyGeometry:
            if type(self._geometry) is not self.parent_object.obj_class.geometry_type:
                raise RuntimeError("Input geometry type {!r} != geometry type of ObjClass {}"
                                   .format(type(self._geometry), self.parent_object.obj_class.geometry_type))

    def to_json(self, key_id_map=None, save_meta=False):
        '''
        The function to_json convert VideoFigure to json format
        :param key_id_map: KeyIdMap class object
        :param save_meta: bool
        :return: VideoFigure in json format
        '''
        data_json = {
            KEY: self.key().hex,
            OBJECT_KEY: self.parent_object.key().hex,
            ApiField.GEOMETRY_TYPE: self.geometry.geometry_name(),
            ApiField.GEOMETRY: self.geometry.to_json()
        }

        if key_id_map is not None:
            item_id = key_id_map.get_figure_id(self.key())
            if item_id is not None:
                data_json[ID] = item_id

            object_id = key_id_map.get_object_id(self.parent_object.key())
            if object_id is not None:
                data_json[OBJECT_ID] = object_id
        if save_meta is True:
            data_json[ApiField.META] = {ApiField.FRAME: self.frame_index}

        self._add_creation_info(data_json)
        return data_json

    @classmethod
    def from_json(cls, data, objects: VideoObjectCollection, frame_index, key_id_map: KeyIdMap = None):
        '''
        The function from_json convert VideoFigure from json format to VideoFigure class object.
        :param data: input VideoFigure in json format
        :param objects: VideoObjectCollection
        :param frame_index: int
        :param key_id_map: KeyIdMap class object
        :return: VideoFigure class object
        '''
        object_id = data.get(OBJECT_ID, None)
        object_key = None
        if OBJECT_KEY in data:
            object_key = uuid.UUID(data[OBJECT_KEY])

        if object_id is None and object_key is None:
            raise RuntimeError("Figure can not be deserialized from json: object_id or object_key are not found")

        if object_key is None:
            if key_id_map is None:
                raise RuntimeError("Figure can not be deserialized: key_id_map is None")
            object_key = key_id_map.get_object_key(object_id)
            if object_key is None:
                raise RuntimeError("Object with id={!r} not found in key_id_map".format(object_id))

        object = objects.get(object_key)
        if object is None:
            raise RuntimeError("Figure can not be deserialized: corresponding object {!r} not found in ObjectsCollection".format(object_key.hex))

        shape_str = data[ApiField.GEOMETRY_TYPE]
        geometry_json = data[ApiField.GEOMETRY]

        shape = GET_GEOMETRY_FROM_STR(shape_str)
        geometry = shape.from_json(geometry_json)

        key = uuid.UUID(data[KEY]) if KEY in data else uuid.uuid4()

        if key_id_map is not None:
            key_id_map.add_figure(key, data.get(ID, None))

        class_id = data.get(CLASS_ID, None)
        labeler_login = data.get(LABELER_LOGIN, None)
        updated_at = data.get(UPDATED_AT, None)
        created_at = data.get(CREATED_AT, None)

        return cls(object, geometry, frame_index, key,
                   class_id=class_id, labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)

    def clone(self, video_object=None, geometry=None, frame_index=None, key=None,
              class_id=None, labeler_login=None, updated_at=None, created_at=None):
        return self.__class__(video_object=take_with_default(video_object, self.parent_object),
                              geometry=take_with_default(geometry, self.geometry),
                              frame_index=take_with_default(frame_index, self.frame_index),
                              key=take_with_default(key, self._key),
                              class_id=take_with_default(class_id, self.class_id),
                              labeler_login=take_with_default(labeler_login, self.labeler_login),
                              updated_at=take_with_default(updated_at, self.updated_at),
                              created_at=take_with_default(created_at, self.created_at)
                              )

    def validate_bounds(self, img_size, _auto_correct=False):
        '''
        The function validate_bounds checks if given image contains a figure. Raise error if figure is out of image bounds
        :param img_size: tuple or list of integers
        :param _auto_correct: bool
        '''
        canvas_rect = Rectangle.from_size(img_size)
        if canvas_rect.contains(self.geometry.to_bbox()) is False:
            raise OutOfImageBoundsExtension("Figure is out of image bounds")

        if _auto_correct is True:
            geometries_after_crop = [cropped_geometry for cropped_geometry in self.geometry.crop(canvas_rect)]
            if len(geometries_after_crop) != 1:
                raise OutOfImageBoundsExtension("Several geometries after crop")
            self._set_geometry_inplace(geometries_after_crop[0])



# coding: utf-8

from supervisely.collection.key_indexed_collection import KeyIndexedCollection
from supervisely.video_annotation.frame import Frame
from supervisely.api.module_api import ApiField


class FrameCollection(KeyIndexedCollection):
    '''
    Collection that stores Frame instances.
    '''
    item_type = Frame

    def to_json(self, key_id_map=None):
        '''
        Converts collection to json serializable list.
        :param key_id_map: KeyIdMap class object
        :return: list of frames in json format
        '''
        return [frame.to_json(key_id_map) for frame in self]

    @classmethod
    def from_json(cls, data, objects, frames_count=None, key_id_map=None):
        '''
        Creates collection from json serializable list
        :param data: list of frames in json format
        :param objects: VideoObjectCollection
        :param frames_count: int
        :param key_id_map: int
        :return: FrameCollection class object
        '''
        frames = [cls.item_type.from_json(frame_json, objects, frames_count, key_id_map) for frame_json in data]
        return cls(frames)

    def __str__(self):
        return 'Frames:\n' + super(FrameCollection, self).__str__()

    @property
    def figures(self):
        '''
        :return: list of figures from all frames in collection
        '''
        figures_array = []
        for frame in self:
            figures_array.extend(frame.figures)
        return figures_array

    def get_figures_and_keys(self, key_id_map):
        '''
        :param key_id_map: KeyIdMap class object
        :return: list of figures from all frames in collection in json format, list of keys from all figures in frames in collection
        '''
        keys = []
        figures_json = []
        for frame in self:
            for figure in frame.figures:
                keys.append(figure.key())
                figure_json = figure.to_json(key_id_map)
                figure_json[ApiField.META] = {ApiField.FRAME: frame.index}
                figures_json.append(figure_json)
        return figures_json, keys

